/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Robby Dermody <robbyd@u20.org>
 * Original code by Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Application that can stream audio data from a URL (via HTTP) in a variety of formats, uses wget.
 * 
 * \ingroup applications
 */
 
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 1.01 $")

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/frame.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/translate.h"
#include "asterisk/options.h"
#include "asterisk/app.h"

#define LOCAL_WGET "/usr/local/bin/wget"
#define WGET "/usr/bin/wget"

static char *tdesc = "Play HTTP streams";

static char *app = "HTTPPlayback";

static char *synopsis = "Play an audio stream over HTTP of the specifed format";

static char *descrip = 
"  HTTPPlayback(location,audioFormat,[timeout])\n"
"Executes wget to get the audio data at the given location, which is a URL.\n"
"User can exit by pressing any key on the dialpad, or by hanging up.\n"
"Return values: 0: No more data to stream; -1: User hung up; "
"Otherwise the ASCII code of the DTMF key the user pressed.\n"
"Valid values of \"audioFmt\" are: "
"\"slinear\", \"ulaw\", \"alaw\", \"g726\", \"adpcm\", \"gsm\", \"ilbc\","
"\"speex_cbr_q3or4\", \"speex_cbr_q5or6\", \"speex_cbr_q7or8\", \"speex_cbr_q9\","
"\"speex_cbr_q10\", \"lpc10\", \"g729a\", \"g7231\".\n"
"timeout, when specified, alters the length of time (in ms) that that application will wait for data on "
"an inactive HTTP stream before giving up and assuming an end of stream situation. If unspecified, this "
"value defaults to 2000 (ms).\n"
"NOTE: With speex, \"speex_cbr_q3or4\" for example means Speex in CBR mode with a "
"quality code of either 3 or 4\n";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static int httpplay(char *url, int fd)
{
	int res;
	int x;
	res = fork();
	if (res < 0) 
		ast_log(LOG_WARNING, "Fork failed\n");
	if (res)
		return res;
	dup2(fd, STDOUT_FILENO);
	for (x=0;x<256;x++) {
		if (x != STDOUT_FILENO)
			close(x);
	}
	/* Execute wget */
	/* Most commonly installed in /usr/bin */
    execl(WGET, "wget", "-q", "-O", "-", url, (char *)NULL);
	/*But many places have it in /usr/local/bin */
    execl(LOCAL_WGET, "wget", "-q", "-O", "-", url, (char *)NULL);
	/* As a last-ditch effort, try to use PATH */
    execlp("wget", "wget", "-q", "-O", "-", url, (char *)NULL);

	ast_log(LOG_WARNING, "Execute of wget failed\n");
	return -1;
}

static int timed_read(int fd, void *data, int datalen, int timeout)
{
	int res;
	struct pollfd fds[1];
	fds[0].fd = fd;
	fds[0].events = POLLIN;
	res = poll(fds, 1, timeout);

	if (res == 0) /*handle a normal stream timeout*/
		return 0;

	if (res < 1) /* poll error */ {
		ast_log(LOG_NOTICE, "Poll errored out with %d\n", res);
		return -1;
	}
	return read(fd, data, datalen);
}

static int httpplay_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	struct localuser *u;
	int fds[2];
	int ms = -1;
	int pid = -1;
	int owriteformat;
	struct timeval next;
	struct ast_frame *f;
	struct myframe {
		struct ast_frame f;
		char offset[AST_FRIENDLY_OFFSET];
		char* frdata;
	} myf;

	char *tmp = NULL;
	/* default to reading in signed linear data */
	int option_audiofmt = AST_FORMAT_SLINEAR;
	int option_timeout = 2000;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(url);
		AST_APP_ARG(audiofmt);
		AST_APP_ARG(timeout);
	);
	
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "HTTP Audio Playback requires an argument (URL)\n");
		return -1;
	}

	LOCAL_USER_ADD(u);

	/* parse args */
	tmp = ast_strdupa(data);
	if (!tmp) {
		ast_log(LOG_ERROR, "Out of memory!\n");
		LOCAL_USER_REMOVE(u);
		return -1;	
	}

	AST_STANDARD_APP_ARGS(args, tmp);

	/* Verify arguments */
	if (!args.url || strncasecmp((char *)args.url, "http://", 7)) {
		ast_log(LOG_ERROR, "Stream URL must start with \"http://\"\n");
		LOCAL_USER_REMOVE(u);
		return -1;	
	}	
 	if (!args.audiofmt) {
		ast_log(LOG_ERROR, "Audio format parameter must be specified!\n");
		LOCAL_USER_REMOVE(u);
		return -1;	
 	}

	/* Parse out audio format argument */
	if (strcasestr(args.audiofmt, "slinear")) 
		option_audiofmt = AST_FORMAT_SLINEAR;
	else if (strcasestr(args.audiofmt, "ulaw")) 
		option_audiofmt = AST_FORMAT_ULAW;
	else if (strcasestr(args.audiofmt, "alaw")) 
		option_audiofmt = AST_FORMAT_ALAW;
	else if (strcasestr(args.audiofmt, "g726")) 
		option_audiofmt = AST_FORMAT_G726;
	else if (strcasestr(args.audiofmt, "adpcm"))
		option_audiofmt = AST_FORMAT_ADPCM;
	else if (strcasestr(args.audiofmt, "gsm")) 
		option_audiofmt = AST_FORMAT_GSM;
	else if (strcasestr(args.audiofmt, "ilbc"))
		option_audiofmt = AST_FORMAT_ILBC;
	else if (strcasestr(args.audiofmt, "speex")) 
		option_audiofmt = AST_FORMAT_SPEEX;
	else if (strcasestr(args.audiofmt, "lpc10"))
		option_audiofmt = AST_FORMAT_LPC10;
	else if (strcasestr(args.audiofmt, "g729a"))
		option_audiofmt = AST_FORMAT_G729A;
	else if (strcasestr(args.audiofmt, "g7231"))
		option_audiofmt = AST_FORMAT_G723_1;
	
	/*choose a buffer size that covers 20ms of data in the specified format
	  20ms of data is 160 samples @ 8HKz, since 8KHz/20ms=160
	  For some formats, 20ms of data doesn't lie on byte boundaries, so we will
	  go up to 30 or 40ms
	  See http://www.openh323.org/docs/bandwidth.html
	  and http://www.voip-info.org/wiki-Codecs
	  and http://www.voip-info.org/wiki/index.php?page=Asterisk+config+codecs.conf
	*/
	if (strcasestr(args.audiofmt, "g7231")) 
		myf.frdata = malloc(16); 
	else if (strcasestr(args.audiofmt, "speex_cbr_q3or4")) /* 8kbps */
		myf.frdata = malloc(20); 
	else if (strcasestr(args.audiofmt, "speex_cbr_q5or6")) /* 11kbps, 40ms */
		myf.frdata = malloc(55); 
	else if (strcasestr(args.audiofmt, "speex_cbr_q7or8")) /* 15kbps, 40ms */
		myf.frdata = malloc(75); 
	else if (strcasestr(args.audiofmt, "speex_cbr_q9")) /* 18.2kbps, 40ms */
		myf.frdata = malloc(91); 
	else if (strcasestr(args.audiofmt, "speex_cbr_q10")) /* 24.6kbps, 40ms */
		myf.frdata = malloc(123); 
	else if (strcasestr(args.audiofmt, "lpc10")) /* 22.5ms (weirdo in the group) */
		myf.frdata = malloc(7);
	else /* ast_codec_get_len() covers the rest of the codecs */
		myf.frdata = malloc(ast_codec_get_len(option_audiofmt, 160));
	
	/* parse out timeout argument (if specified) */
	if(args.timeout) {
		option_timeout = atoi(args.timeout);
	}

	if (pipe(fds)) {
		ast_log(LOG_WARNING, "Unable to create pipe\n");
		LOCAL_USER_REMOVE(u);
		free(myf.frdata);
		return -1;
	}
	
	ast_stopstream(chan);

	owriteformat = chan->writeformat;
	res = ast_set_write_format(chan, option_audiofmt);
	if (res < 0) {
		ast_log(LOG_WARNING, "Unable to set write format to %i\n", option_audiofmt);
		LOCAL_USER_REMOVE(u);
		free(myf.frdata);
		return -1;
	}
	
	ast_log(LOG_DEBUG, "Playing URL \"%s\" with file format \"%s\" (%i). Timeout is %i ms\n",
		    args.url, args.audiofmt, option_audiofmt, option_timeout);
	
	res = httpplay((char *)args.url, fds[1]);
	/* Wait 1000 ms first */
	next = ast_tvnow();
	next.tv_sec += 1;
	if (res >= 0) {
		pid = res;
		for (;;) {
			ms = ast_tvdiff_ms(next, ast_tvnow());
			if (ms <= 0) {
				res = timed_read(fds[0], myf.frdata, sizeof(myf.frdata), option_timeout);
				if (res > 0) {
					myf.f.frametype = AST_FRAME_VOICE;
					myf.f.subclass = option_audiofmt;
					myf.f.datalen = res;
					myf.f.mallocd = 0;
					myf.f.offset = AST_FRIENDLY_OFFSET;
					myf.f.src = __PRETTY_FUNCTION__;
					myf.f.delivery.tv_sec = 0;
					myf.f.delivery.tv_usec = 0;
					myf.f.data = myf.frdata;
					myf.f.samples = ast_codec_get_samples(&myf.f);
					if (ast_write(chan, &myf.f) < 0) {
						res = -1;
						break;
					}
				} else {
					ast_log(LOG_DEBUG, "No more data to read\n");
					res = 0;
					break;
				}
				next = ast_tvadd(next, ast_samp2tv(myf.f.samples, 8000)); /* pause a second? */
			} else {
				ms = ast_waitfor(chan, ms);
				if (ms < 0) {
					ast_log(LOG_DEBUG, "Hangup detected\n");
					res = -1;
					break;
				}
				if (ms) {
					f = ast_read(chan);
					if (!f) {
						ast_log(LOG_DEBUG, "Null frame == hangup() detected\n");
						res = -1;
						break;
					}
					if (f->frametype == AST_FRAME_DTMF) {
						ast_log(LOG_DEBUG, "User pressed a key: %c\n", f->subclass);
						res = f->subclass; /* return the key that the user pressed (as an int) */
						ast_frfree(f);
						break;
					}
					ast_frfree(f);
				} 
			}
		}
	}
	close(fds[0]);
	close(fds[1]);
	
	if (pid > -1)
		kill(pid, SIGKILL);
	if (!res && owriteformat)
		ast_set_write_format(chan, owriteformat);

	LOCAL_USER_REMOVE(u);
	
	free(myf.frdata);
	return res;
}

int unload_module(void)
{
	int res;

	res = ast_unregister_application(app);

	STANDARD_HANGUP_LOCALUSERS;
	
	return res;
}

int load_module(void)
{
	return ast_register_application(app, httpplay_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
